@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
@@ -10,31 +10,33 @@ from functools import partial
 from gettext import gettext as _
 from weakref import WeakValueDictionary
 
+from .child import cached_process_data
 from .cli import create_opts, parse_args
 from .conf.utils import to_cmdline
 from .config import initial_window_size_func, prepare_config_file_for_editing
 from .config_data import MINIMUM_FONT_SIZE
 from .constants import (
-    appname, config_dir, kitty_exe, set_boss, supports_primary_selection
+    appname, config_dir, is_macos, kitty_exe, set_boss,
+    supports_primary_selection
 )
 from .fast_data_types import (
     ChildMonitor, background_opacity_of, change_background_opacity,
     change_os_window_state, create_os_window, current_os_window,
-    destroy_global_data, get_clipboard_string, glfw_post_empty_event,
-    global_font_size, mark_os_window_for_close, os_window_font_size,
-    patch_global_colors, set_clipboard_string, set_in_sequence_mode,
-    toggle_fullscreen
+    destroy_global_data, get_clipboard_string, global_font_size,
+    mark_os_window_for_close, os_window_font_size, patch_global_colors,
+    set_clipboard_string, set_in_sequence_mode, toggle_fullscreen,
+    toggle_maximized
 )
 from .keys import get_shortcut, shortcut_matches
-from .layout import set_draw_minimal_borders
+from .layout import set_draw_borders_options
 from .remote_control import handle_cmd
 from .rgb import Color, color_from_int
-from .session import create_session
+from .session import create_sessions
 from .tabs import SpecialWindow, SpecialWindowInstance, TabManager
 from .utils import (
-    get_editor, get_primary_selection, log_error, open_url, parse_address_spec,
-    remove_socket_file, safe_print, set_primary_selection, single_instance,
-    startup_notification_handler
+    func_name, get_editor, get_primary_selection, is_path_in_temp_dir,
+    log_error, open_url, parse_address_spec, remove_socket_file, safe_print,
+    set_primary_selection, single_instance, startup_notification_handler
 )
 
 
@@ -103,9 +105,12 @@ class DumpCommands:  # {{{
 class Boss:
 
     def __init__(self, os_window_id, opts, args, cached_values, new_os_window_trigger):
-        set_draw_minimal_borders(opts)
+        set_draw_borders_options(opts)
+        self.clipboard_buffers = {}
+        self.update_check_process = None
         self.window_id_map = WeakValueDictionary()
         self.startup_colors = {k: opts[k] for k in opts if isinstance(opts[k], Color)}
+        self.startup_cursor_text_color = opts.cursor_text_color
         self.pending_sequences = None
         self.cached_values = cached_values
         self.os_window_map = {}
@@ -124,20 +129,25 @@ class Boss:
         )
         set_boss(self)
         self.opts, self.args = opts, args
-        startup_session = create_session(opts, args, default_session=opts.startup_session)
+        startup_sessions = create_sessions(opts, args, default_session=opts.startup_session)
         self.keymap = self.opts.keymap.copy()
         if new_os_window_trigger is not None:
             self.keymap.pop(new_os_window_trigger, None)
-        self.add_os_window(startup_session, os_window_id=os_window_id)
-        if args.start_as != 'normal':
-            if args.start_as == 'fullscreen':
-                self.toggle_fullscreen()
-            else:
-                change_os_window_state(args.start_as)
+        for startup_session in startup_sessions:
+            os_window_id = self.add_os_window(startup_session, os_window_id=os_window_id)
+            if args.start_as != 'normal':
+                if args.start_as == 'fullscreen':
+                    self.toggle_fullscreen()
+                else:
+                    change_os_window_state(args.start_as)
+            os_window_id = None
+        if is_macos:
+            from .fast_data_types import cocoa_set_notification_activated_callback
+            cocoa_set_notification_activated_callback(self.notification_activated)
 
     def add_os_window(self, startup_session, os_window_id=None, wclass=None, wname=None, opts_for_size=None, startup_id=None):
         if os_window_id is None:
-            opts_for_size = opts_for_size or self.opts
+            opts_for_size = opts_for_size or startup_session.os_window_size or self.opts
             cls = wclass or self.args.cls or appname
             with startup_notification_handler(do_notify=startup_id is not None, startup_id=startup_id) as pre_show_callback:
                 os_window_id = create_os_window(
@@ -149,14 +159,15 @@ class Boss:
         return os_window_id
 
     def list_os_windows(self):
-        active_tab, active_window = self.active_tab, self.active_window
-        active_tab_manager = self.active_tab_manager
-        for os_window_id, tm in self.os_window_map.items():
-            yield {
-                'id': os_window_id,
-                'is_focused': tm is active_tab_manager,
-                'tabs': list(tm.list_tabs(active_tab, active_window)),
-            }
+        with cached_process_data():
+            active_tab, active_window = self.active_tab, self.active_window
+            active_tab_manager = self.active_tab_manager
+            for os_window_id, tm in self.os_window_map.items():
+                yield {
+                    'id': os_window_id,
+                    'is_focused': tm is active_tab_manager,
+                    'tabs': list(tm.list_tabs(active_tab, active_window)),
+                }
 
     @property
     def all_tab_managers(self):
@@ -238,7 +249,7 @@ class Boss:
             sw = args
         else:
             sw = self.args_to_special_window(args, cwd_from) if args else None
-        startup_session = create_session(self.opts, special_window=sw, cwd_from=cwd_from)
+        startup_session = next(create_sessions(self.opts, special_window=sw, cwd_from=cwd_from))
         return self.add_os_window(startup_session)
 
     def new_os_window(self, *args):
@@ -253,9 +264,13 @@ class Boss:
 
     def new_os_window_with_cwd(self, *args):
         w = self.active_window_for_cwd
-        cwd_from = w.child.pid if w is not None else None
+        cwd_from = w.child.pid_for_cwd if w is not None else None
         self._new_os_window(args, cwd_from)
 
+    def new_os_window_with_wd(self, wd):
+        special_window = SpecialWindow(None, cwd=wd)
+        self._new_os_window(special_window)
+
     def add_child(self, window):
         self.child_monitor.add_child(window.id, window.child.pid, window.child.child_fd, window.screen)
         self.window_id_map[window.id] = window
@@ -292,10 +307,10 @@ class Boss:
                 opts = create_opts(args)
                 if not os.path.isabs(args.directory):
                     args.directory = os.path.join(msg['cwd'], args.directory)
-                session = create_session(opts, args, respect_cwd=True)
-                os_window_id = self.add_os_window(session, wclass=args.cls, wname=args.name, opts_for_size=opts, startup_id=startup_id)
-                if msg.get('notify_on_os_window_death'):
-                    self.os_window_death_actions[os_window_id] = partial(self.notify_on_os_window_death, msg['notify_on_os_window_death'])
+                for session in create_sessions(opts, args, respect_cwd=True):
+                    os_window_id = self.add_os_window(session, wclass=args.cls, wname=args.name, opts_for_size=opts, startup_id=startup_id)
+                    if msg.get('notify_on_os_window_death'):
+                        self.os_window_death_actions[os_window_id] = partial(self.notify_on_os_window_death, msg['notify_on_os_window_death'])
             else:
                 log_error('Unknown message received from peer, ignoring')
 
@@ -332,7 +347,6 @@ class Boss:
             if len(tm) == 0:
                 if not self.shutting_down:
                     mark_os_window_for_close(os_window_id)
-                    glfw_post_empty_event()
 
     def close_window(self, window=None):
         if window is None:
@@ -348,10 +362,17 @@ class Boss:
     def toggle_fullscreen(self):
         toggle_fullscreen()
 
+    def toggle_maximized(self):
+        toggle_maximized()
+
     def start(self):
         if not getattr(self, 'io_thread_started', False):
             self.child_monitor.start()
             self.io_thread_started = True
+        if self.opts.update_check_interval > 0 and not hasattr(self, 'update_check_started'):
+            from .update_check import run_update_check
+            run_update_check(self.opts.update_check_interval * 60 * 60)
+            self.update_check_started = True
 
     def activate_tab_at(self, os_window_id, x):
         tm = self.os_window_map.get(os_window_id)
@@ -560,7 +581,7 @@ class Boss:
             f = getattr(self, key_action.func, None)
             if f is not None:
                 if self.args.debug_keyboard:
-                    print('Keypress matched action:', f.__name__)
+                    print('Keypress matched action:', func_name(f))
                 passthrough = f(*key_action.args)
                 if passthrough is not True:
                     return True
@@ -575,7 +596,7 @@ class Boss:
             if f is not None:
                 passthrough = f(*key_action.args)
                 if self.args.debug_keyboard:
-                    print('Keypress matched action:', f.__name__)
+                    print('Keypress matched action:', func_name(f))
                 if passthrough is not True:
                     return True
         return False
@@ -725,7 +746,7 @@ class Boss:
             self._run_kitten('ask', args)
 
     def show_error(self, title, msg):
-        self._run_kitten('show_error', ['--title', title], input_data=msg)
+        self._run_kitten('show_error', args=['--title', title], input_data=msg)
 
     def do_set_tab_title(self, title, tab_id):
         tm = self.active_tab_manager
@@ -767,6 +788,8 @@ class Boss:
     def destroy(self):
         self.shutting_down = True
         self.child_monitor.shutdown_monitor()
+        self.set_update_check_process()
+        self.update_check_process = None
         del self.child_monitor
         for tm in self.os_window_map.values():
             tm.destroy()
@@ -794,7 +817,29 @@ class Boss:
             if text:
                 set_primary_selection(text)
                 if self.opts.copy_on_select:
+                    self.copy_to_buffer(self.opts.copy_on_select)
+
+    def copy_to_buffer(self, buffer_name):
+        w = self.active_window
+        if w is not None and not w.destroyed:
+            text = w.text_for_selection()
+            if text:
+                if buffer_name == 'clipboard':
                     set_clipboard_string(text)
+                elif buffer_name == 'primary':
+                    set_primary_selection(text)
+                else:
+                    self.clipboard_buffers[buffer_name] = text
+
+    def paste_from_buffer(self, buffer_name):
+        if buffer_name == 'clipboard':
+            text = get_clipboard_string()
+        elif buffer_name == 'primary':
+            text = get_primary_selection()
+        else:
+            text = self.clipboard_buffers.get(buffer_name)
+        if text:
+            self.paste_to_active_window(text)
 
     def goto_tab(self, tab_num):
         tm = self.active_tab_manager
@@ -852,7 +897,7 @@ class Boss:
     def pipe(self, source, dest, exe, *args):
         cmd = [exe] + list(args)
         window = self.active_window
-        cwd_from = window.child.pid if window else None
+        cwd_from = window.child.pid_for_cwd if window else None
 
         def create_window():
             return self.special_window_for_cmd(
@@ -920,9 +965,13 @@ class Boss:
 
     def new_tab_with_cwd(self, *args):
         w = self.active_window_for_cwd
-        cwd_from = w.child.pid if w is not None else None
+        cwd_from = w.child.pid_for_cwd if w is not None else None
         self._create_tab(args, cwd_from=cwd_from)
 
+    def new_tab_with_wd(self, wd):
+        special_window = SpecialWindow(None, cwd=wd)
+        self._new_tab(special_window)
+
     def _new_window(self, args, cwd_from=None):
         tab = self.active_tab
         if tab is not None:
@@ -938,7 +987,7 @@ class Boss:
         w = self.active_window_for_cwd
         if w is None:
             return self.new_window(*args)
-        cwd_from = w.child.pid if w is not None else None
+        cwd_from = w.child.pid_for_cwd if w is not None else None
         self._new_window(args, cwd_from=cwd_from)
 
     def move_tab_forward(self):
@@ -951,11 +1000,83 @@ class Boss:
         if tm is not None:
             tm.move_tab(-1)
 
-    def patch_colors(self, spec, configured=False):
+    def disable_ligatures_in(self, where, strategy):
+        if isinstance(where, str):
+            windows = ()
+            if where == 'active':
+                if self.active_window is not None:
+                    windows = (self.active_window,)
+            elif where == 'all':
+                windows = self.all_windows
+            elif where == 'tab':
+                if self.active_tab is not None:
+                    windows = tuple(self.active_tab)
+        else:
+            windows = where
+        for window in windows:
+            window.screen.disable_ligatures = strategy
+            window.refresh()
+
+    def patch_colors(self, spec, cursor_text_color, configured=False):
         if configured:
             for k, v in spec.items():
                 if hasattr(self.opts, k):
                     setattr(self.opts, k, color_from_int(v))
+            if cursor_text_color is not False:
+                if isinstance(cursor_text_color, int):
+                    cursor_text_color = color_from_int(cursor_text_color)
+                self.opts.cursor_text_color = cursor_text_color
         for tm in self.all_tab_managers:
             tm.tab_bar.patch_colors(spec)
         patch_global_colors(spec, configured)
+
+    def safe_delete_temp_file(self, path):
+        if is_path_in_temp_dir(path):
+            try:
+                os.remove(path)
+            except FileNotFoundError:
+                pass
+
+    def set_update_check_process(self, process=None):
+        if self.update_check_process is not None:
+            try:
+                if self.update_check_process.poll() is None:
+                    self.update_check_process.kill()
+            except Exception:
+                pass
+        self.update_check_process = process
+
+    def on_monitored_pid_death(self, pid, exit_status):
+        update_check_process = getattr(self, 'update_check_process', None)
+        if update_check_process is not None and pid == update_check_process.pid:
+            self.update_check_process = None
+            from .update_check import process_current_release
+            try:
+                raw = update_check_process.stdout.read().decode('utf-8')
+            except Exception as e:
+                log_error('Failed to read data from update check process, with error: {}'.format(e))
+            else:
+                try:
+                    process_current_release(raw)
+                except Exception as e:
+                    log_error('Failed to process update check data {!r}, with error: {}'.format(raw, e))
+
+    def notification_activated(self, identifier):
+        if identifier == 'new-version':
+            from .update_check import notification_activated
+            notification_activated()
+
+    def dbus_notification_callback(self, activated, *args):
+        from .notify import dbus_notification_created, dbus_notification_activated
+        if activated:
+            dbus_notification_activated(*args)
+        else:
+            dbus_notification_created(*args)
+
+    def show_bad_config_lines(self, bad_lines):
+
+        def format_bad_line(bad_line):
+            return '{}:{} in line: {}\n'.format(bad_line.number, bad_line.exception, bad_line.line)
+
+        msg = '\n'.join(map(format_bad_line, bad_lines)).rstrip()
+        self.show_error(_('Errors in kitty.conf'), msg)
